import sys

from ..forms import IssueFormCleanedData
from ._base import BaseBackend


class Backend(BaseBackend):
    def create_ticket(self, cleaned_data: IssueFormCleanedData) -> bool:
        from ..config import CONFIG

        if "screenshot" in cleaned_data:
            screenshot_url = "<screenshot>"
        else:
            screenshot_url = "<screenshot not provided>"

        data = {
            "title": cleaned_data["title"],
            "type": cleaned_data["type"],
            "description": self.get_description({**cleaned_data, "screenshot_url": screenshot_url}),
        }

        data = {
            "labels": CONFIG.ANNOTATIONS["get_labels"](self.request, [cleaned_data.get("type", "issue")]),
            **data,
        }
        for k, v in data.items():
            sys.stdout.write(f"{k}: {v}")
        return True



import base64

import markdown
from django.core.mail import EmailMessage

from ..forms import IssueFormCleanedData
from ._base import BaseBackend


class Backend(BaseBackend):
    def create_ticket(self, cleaned_data: IssueFormCleanedData) -> bool:
        screenshot = cleaned_data["screenshot"]
        description = self.get_description({**cleaned_data, "screenshot_url": ""})
        html_output = markdown.markdown(description)
        email = EmailMessage(
            subject=f"[{cleaned_data['type']}] {cleaned_data['title']}",
            body=description,
            from_email=self.get_option("SENDER"),
            to=self.get_option("RECIPIENTS"),
        )
        email.attach("body", html_output, "text/html")
        if screenshot:
            base64_string = screenshot.split("base64,")[1]
            decoded_data = base64.b64decode(base64_string)
            email.attach("screenshot.png", decoded_data, "image/png")

        email.send()
        return True



import base64
import datetime

import github
from django.utils.text import slugify
from github import Github

from ..config import CONFIG
from ..forms import IssueFormCleanedData
from .github import Backend as GithubBackend


class Backend(GithubBackend):
    screenshot_supported = True

    def create_ticket(self, cleaned_data: IssueFormCleanedData) -> bool:
        screenshot_path = self.get_option("SCREENSHOT_REPO_PATH")
        screenshot_url = ""

        g = Github(auth=github.Auth.Token(self.get_option("API_TOKEN")))
        repo = g.get_repo(self.get_option("PROJECT"))

        if cleaned_data["screenshot"]:
            screenshot = cleaned_data["screenshot"].split(",")[1]
            screenshot_data = base64.b64decode(screenshot)
            timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            slug = slugify(cleaned_data["title"])
            filename = f"{slug}_{timestamp}.png"
            path = f"{screenshot_path}{filename}"

            repo.create_file(
                path=path,
                message=f"Add screenshot for issue: {cleaned_data['title']}",
                content=screenshot_data,
                branch=self.get_option("SCREENSHOT_BRANCH"),
            )
            url = repo.get_contents(path).download_url  # type: ignore[union-attr]
            screenshot_url = f"![Screenshot]({url})"

        description = self.get_description({**cleaned_data, "screenshot_url": screenshot_url})

        repo.create_issue(
            title=cleaned_data["title"],
            body=description,
            labels=CONFIG.ANNOTATIONS["get_labels"](self.request, [cleaned_data["type"]]),
        )
        return True



from typing import TYPE_CHECKING, Any

from ..forms import IssueFormCleanedData
from ._base import BaseBackend

if TYPE_CHECKING:
    from django.http import HttpRequest


class Backend(BaseBackend):
    tickets: list[dict[str, Any]] = []

    def __init__(self, request: "HttpRequest") -> None:
        super().__init__(request)
        Backend.tickets = []

    def create_ticket(self, cleaned_data: IssueFormCleanedData) -> bool:
        from ..config import CONFIG

        screenshot_url = ""
        if cleaned_data["screenshot"]:
            screenshot_url = "<url>"

        data = {
            "title": cleaned_data["title"],
            "type": cleaned_data["type"],
            "description": self.get_description({**cleaned_data, "screenshot_url": screenshot_url}),
            "labels": CONFIG.ANNOTATIONS["get_labels"](self.request, [cleaned_data.get("type", "issue")]),
            "screenshot_url": screenshot_url,
            "screenshot": cleaned_data["screenshot"],
        }

        self.tickets.append(data)
        return True



import base64
import io
import json

import markdown
import requests
from django.utils.text import slugify
from requests.auth import HTTPBasicAuth

from ..forms import IssueFormCleanedData
from ._base import BaseBackend


class Backend(BaseBackend):
    screenshot_supported = True

    def create_ticket(self, cleaned_data: IssueFormCleanedData) -> bool:
        server_url = self.get_option("SERVER_URL", "https://dev.azure.com")
        project = self.get_option("PROJECT")  # encode spaces if any
        pat = self.get_option("TOKEN")

        work_item_type = cleaned_data["type"]

        url = f"{server_url}/{project}/_apis/wit/workitems/${work_item_type}?api-version=7.1-preview.3"

        headers = {"Content-Type": "application/json-patch+json"}
        screenshot_url: str = ""
        md_url = ""
        screenshot_data = cleaned_data.get("screenshot")
        if screenshot_data:
            slug = slugify(cleaned_data["title"])
            filename = f"{slug}.png"

            image_data = base64.b64decode(screenshot_data.split(",")[1])
            stream = io.BytesIO(image_data)
            attachment_url = (
                f"{server_url}/{project}/_apis/wit/attachments?fileName={filename}&api-version=7.1-preview.3"
            )
            response = requests.post(
                attachment_url,
                data=stream,
                auth=("", pat),
                headers={"Content-Type": "application/octet-stream"},
                timeout=10,
            )
            response.raise_for_status()
            screenshot_url = response.json()["url"]
            md_url = f"![aa]({response.json()['url']})"

        description = self.get_description({**cleaned_data, "screenshot_url": md_url})
        data_issue = [
            {"op": "add", "path": "/fields/System.Title", "from": None, "value": cleaned_data["title"]},
            {
                "op": "add",
                "path": "/fields/System.Description",
                "from": None,
                "value": markdown.markdown(description),
            },
        ]

        response = requests.post(
            url, headers=headers, auth=HTTPBasicAuth("", pat), data=json.dumps(data_issue), timeout=10
        )
        if response.status_code == 200:  # noqa: PLR2004
            issue = response.json()
            work_item_id = issue["id"]
            if screenshot_data and screenshot_url:
                url = f"https://dev.azure.com/{project}/_apis/wit/workitems/{work_item_id}?api-version=7.1-preview.3"
                data_attachment = [
                    {
                        "op": "add",
                        "path": "/relations/-",
                        "value": {
                            "rel": "AttachedFile",
                            "url": screenshot_url,
                            "attributes": {"comment": "Adding screenshot as evidence"},
                        },
                    }
                ]
                response = requests.patch(
                    url, headers=headers, auth=HTTPBasicAuth("", pat), data=json.dumps(data_attachment), timeout=10
                )
                response.raise_for_status()
            return True
        return False



from collections.abc import Iterable
from typing import Any

from django.apps import AppConfig
from django.conf import settings
from django.core.checks import CheckMessage, Error, register
from django.utils.module_loading import import_string

from .backends._base import BaseBackend


@register("issues")
def check_issues_settings(app_configs: Iterable[AppConfig] | None, **kwargs: Any) -> list[CheckMessage]:
    errors: list[CheckMessage] = []
    override_issues_settings = getattr(settings, "ISSUES", {})

    if override_issues_settings:
        if not isinstance(override_issues_settings, dict):
            errors.append(
                Error(
                    "'ISSUES' setting must be a dictionary.",
                    id="issues.E001",
                )
            )
            return errors  # Stop here if ISSUES is not a dict

        # Check BACKEND
        backend_path = override_issues_settings.get("BACKEND", "")
        try:
            backend_class = import_string(backend_path)
            if not issubclass(backend_class, BaseBackend):
                errors.append(
                    Error(
                        f"'{backend_path}' must be a subclass of 'issues.backends._base.BaseBackend'.",
                        id="issues.E003",
                    )
                )
        except ImportError:
            errors.append(
                Error(
                    f"Could not import backend '{backend_path}'.",
                    id="issues.E004",
                )
            )

        # Check RENDERER
        renderer = override_issues_settings.get("RENDERER")
        allowed_renderers = ["html2canvas", "dom-to-image"]
        if renderer is not None and renderer not in allowed_renderers:
            errors.append(
                Error(
                    f"'{renderer}' is not a valid renderer. Must be one of {allowed_renderers} or None.",
                    id="issues.E005",
                )
            )

        # Check OPTIONS
        if "OPTIONS" in override_issues_settings and not isinstance(override_issues_settings["OPTIONS"], dict):
            errors.append(
                Error(
                    "'OPTIONS' in 'ISSUES' setting must be a dictionary.",
                    id="issues.E006",
                )
            )

    return errors



from collections import ChainMap
from typing import Any

from django.conf import settings
from django.core.signals import setting_changed
from django.dispatch import receiver
from django.utils.module_loading import import_string

ISSUE_TEMPLATE = """
- User: {user}
- Url: {url}
- Version: {version}
- UserAgent: {user_agent}
- Remote IP: {remote_ip}

---
{description}

{screenshot_url}
"""


class IssuesConfig:
    _ANNOTATIONS: dict[str, str] = {
        "get_client_ip": "issues.utils.get_client_ip",
        "get_extra_info": "issues.utils.get_extra_info",
        "get_labels": "issues.utils.get_labels",
        "get_user": "issues.utils.get_user",
        "get_user_agent": "issues.utils.get_user_agent",
        "get_version": "issues.utils.get_version",
    }
    _DEFAULTS = {
        "BACKEND": "issues.backends.console.Backend",
        "RENDERER": "html2canvas",  # or rasterizeHTML or dom-to-image
        "ISSUE_TEMPLATE": ISSUE_TEMPLATE,
        "TYPES": ("bug", "enhancement", "feature", "suggestion"),
        "OPTIONS": {},
    }

    def __init__(self) -> None:
        self._overrides = getattr(settings, "ISSUES", {})
        ann = {"ANNOTATIONS": ChainMap(self._overrides.get("ANNOTATIONS", {}), self._ANNOTATIONS)}
        self._parsed = ChainMap(ann, self._overrides, self._DEFAULTS)
        self._cached: dict[str, Any] = {}

    def __repr__(self) -> str:
        return str(self._parsed)

    def __getattr__(self, name: str) -> Any:
        if name not in self._parsed:
            raise AttributeError(f" 'IssuesConfig' object has no attribute '{name}'")
        if name not in self._cached:
            if name == "BACKEND":
                self._cached[name] = import_string(self._parsed["BACKEND"])
            elif name == "ANNOTATIONS":
                self._cached["ANNOTATIONS"] = {}
                for k, v in self._parsed["ANNOTATIONS"].items():
                    if k not in self._ANNOTATIONS:
                        raise AttributeError(f"Misspelled or unknown annotation '{k}'")
                    if isinstance(v, str) and v:
                        self._cached["ANNOTATIONS"][k] = import_string(v)
                    elif callable(v):
                        self._cached["ANNOTATIONS"][k] = v
                    else:
                        raise AttributeError(f" 'IssuesConfig' object has no annotation '{k}'")
            else:
                self._cached[name] = self._parsed[name]
        return self._cached[name]


CONFIG = IssuesConfig()


@receiver(setting_changed)
def reload_issues_config(sender: Any, setting: str, **kwargs: Any) -> None:
    global CONFIG  # noqa: PLW0603
    if setting == "ISSUES":
        CONFIG = IssuesConfig()



import base64
import io
from typing import TYPE_CHECKING

import requests

from ._base import BaseBackend

if TYPE_CHECKING:
    from ..forms import IssueFormCleanedData


class Backend(BaseBackend):
    screenshot_supported = True

    def get_issue_choices(self) -> list[tuple[str, str]]:
        from issues.config import CONFIG

        return list(CONFIG.TYPES.items())

    def create_ticket(self, cleaned_data: "IssueFormCleanedData") -> bool:
        description = self.get_description({**cleaned_data, "screenshot_url": ""})

        api_url = self.get_option("API_URL", "https://api.taiga.io/api/v1")
        api_token = self.get_option("API_TOKEN")
        project = int(self.get_option("PROJECT_ID"))
        tags = self.get_option("TAGS", [])

        screenshot = cleaned_data["screenshot"]

        headers = {
            "Authorization": f"Application {api_token}",
            "Content-Type": "application/json",
        }
        data = {
            "project": project,
            "subject": cleaned_data["title"],
            "description": description,
            "type": cleaned_data["type"],
            "tags": tags,
        }
        response = requests.post(f"{api_url}/issues", headers=headers, json=data, timeout=10)
        response.raise_for_status()
        issue_data = response.json()

        if screenshot:
            headers.pop("Content-Type")
            image_data = base64.b64decode(screenshot.split(",")[1])
            stream = io.BytesIO(image_data)
            files = {"attached_file": ("screenshot.png", stream, "image/png")}
            form_data = {"project": project, "object_id": issue_data["id"]}
            attachment_response = requests.post(
                f"{api_url}/issues/attachments", headers=headers, data=form_data, files=files, timeout=10
            )
            attachment_response.raise_for_status()

        return True



from django.apps import AppConfig


class IssuesConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "issues"

    def ready(self) -> None:
        from . import checks  # noqa



from typing import TYPE_CHECKING, Any, TypedDict

from issues.exceptions import IssueError
from issues.forms import IssueFormCleanedData

if TYPE_CHECKING:
    from django.http import HttpRequest

not_provided = object()


class IssueData(TypedDict):
    title: str
    type: str
    description: str


class BaseBackend:
    screenshot_supported: bool = True

    def __init__(self, request: "HttpRequest") -> None:
        self.request = request

    def get_issue_choices(self) -> list[tuple[str, str]]:
        from issues.config import CONFIG

        return [(issue_type, issue_type) for issue_type in CONFIG.TYPES]

    def get_option(self, name: str, default: Any = not_provided) -> Any:
        from issues.config import CONFIG

        try:
            return CONFIG.OPTIONS[name]
        except KeyError as e:
            if default == not_provided:
                raise IssueError("Issues backend Improperly configured.") from e
            return default

    def get_context(self) -> dict[str, Any]:
        from issues.config import CONFIG

        data = {
            "extras": {},
            "url": self.request.META.get("HTTP_REFERER", "N/A"),
            "user": CONFIG.ANNOTATIONS["get_user"](self.request),
            "user_agent": CONFIG.ANNOTATIONS["get_user_agent"](self.request),
            "version": CONFIG.ANNOTATIONS["get_version"](self.request),
            "remote_ip": CONFIG.ANNOTATIONS["get_client_ip"](self.request),
        }
        data["extras"] = CONFIG.ANNOTATIONS["get_extra_info"](self.request, data)
        return data

    def get_description(self, parameters: dict[str, Any]) -> str:
        from issues.config import CONFIG

        template: str = CONFIG.ISSUE_TEMPLATE
        return template.format(**self.get_context(), **parameters)

    def create_ticket(self, cleaned_data: IssueFormCleanedData) -> bool:
        raise NotImplementedError(f"{self.__class__.__name__} does not implement create_ticket")



from typing import TYPE_CHECKING, Any

from django.http import HttpRequest, HttpResponseRedirect

if TYPE_CHECKING:
    from issues.types import AuthenticatedHttpRequest


class HttpResponseRedirectToReferrer(HttpResponseRedirect):
    def __init__(self, request: HttpRequest, *args: tuple[Any], **kwargs: dict[str, Any]) -> None:
        redirect_to = request.META.get("HTTP_REFERER", "/")
        super().__init__(redirect_to, False, *args, **kwargs)


def get_extra_info(request: HttpRequest, data: dict[str, Any]) -> dict[str, Any]:
    return {}


def get_client_ip(request: HttpRequest) -> str:
    for x in [
        "HTTP_X_ORIGINAL_FORWARDED_FOR",
        "HTTP_X_FORWARDED_FOR",
        "HTTP_X_REAL_IP",
        "REMOTE_ADDR",
    ]:
        ip = request.META.get(x)
        if ip:
            return str(ip.split(",")[0].strip())
    return ""


def get_labels(request: HttpRequest, original: list[str]) -> list[str]:
    return original


def get_user_agent(request: HttpRequest) -> str:
    ua_string = request.META.get("HTTP_USER_AGENT", "N/A")
    return str(ua_string)


def get_user(request: "AuthenticatedHttpRequest") -> str:
    if request.user.is_authenticated:
        return request.user.email
    return "N/A"


def get_version(request: HttpRequest) -> str:
    return "N/A"



from typing import TYPE_CHECKING

from ._base import BaseBackend

if TYPE_CHECKING:
    from django.http import HttpRequest


def get_backend(request: "HttpRequest") -> BaseBackend:
    from ..config import CONFIG

    backend_class: type[BaseBackend] = CONFIG.BACKEND
    return backend_class(request)


__all__ = ["get_backend"]



from typing import TYPE_CHECKING, Any, TypedDict

from django import forms

if TYPE_CHECKING:
    from issues.backends._base import BaseBackend


class IssueFormCleanedData(TypedDict):
    type: str
    title: str
    description: str
    add_screenshot: bool
    screenshot: str


class IssueForm(forms.Form):
    type = forms.ChoiceField(
        choices=(
            ("bug", "bug"),
            ("enhancement", "enhancement"),
            ("feature", "feature"),
            ("suggestion", "suggestion"),
        )
    )
    title = forms.CharField(widget=forms.TextInput({"placeholder": "issue title", "autofocus": "autofocus"}))
    description = forms.CharField(widget=forms.Textarea({"placeholder": "description"}))
    add_screenshot = forms.BooleanField(required=False, widget=forms.CheckboxInput, initial=True)
    screenshot = forms.CharField(widget=forms.HiddenInput, required=False)

    def __init__(self, *args: Any, backend: "BaseBackend", **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.fields["type"].choices = backend.get_issue_choices()  # type: ignore[attr-defined]



import base64

import gitlab

from ..forms import IssueFormCleanedData
from ._base import BaseBackend


class Backend(BaseBackend):
    def create_ticket(self, cleaned_data: IssueFormCleanedData) -> bool:
        from ..config import CONFIG

        screenshot = cleaned_data["screenshot"]

        gl = gitlab.Gitlab(private_token=self.get_option("API_TOKEN"))
        gl.auth()
        project = gl.projects.get(self.get_option("PROJECT"))
        screenshot_url = ""
        if screenshot:
            snap = project.upload(
                "screenshot.png",
                base64.b64decode(str(cleaned_data["screenshot"][21:])),
            )
            screenshot_url = snap.get("markdown")  # type: ignore[assignment]
        description = self.get_description({**cleaned_data, "screenshot_url": screenshot_url})
        project.issues.create(
            {
                "title": cleaned_data["title"],
                "description": description,
                "labels": CONFIG.ANNOTATIONS["get_labels"](self.request, [cleaned_data["type"]]),
            }
        )
        return True



import github
from github import Github

from ..config import CONFIG
from ..forms import IssueFormCleanedData
from ._base import BaseBackend


class Backend(BaseBackend):
    screenshot_supported = False

    def create_ticket(self, cleaned_data: IssueFormCleanedData) -> bool:
        description = self.get_description({**cleaned_data, "screenshot_url": ""})

        g = Github(auth=github.Auth.Token(self.get_option("API_TOKEN")))
        repo = g.get_repo(self.get_option("PROJECT"))
        repo.create_issue(
            title=cleaned_data["title"],
            body=description,
            labels=CONFIG.ANNOTATIONS["get_labels"](self.request, [cleaned_data["type"]]),
        )
        return True



from django import template
from django.conf import settings
from django.templatetags.static import static
from django.urls import reverse
from django.utils.safestring import mark_safe

register = template.Library()


@register.simple_tag
def issues_tags() -> str:
    from issues.config import CONFIG

    static_url = static("/")
    url = reverse("issues:create")

    if settings.DEBUG:
        suffix = ""
    else:
        suffix = ".min"
    engine = CONFIG.RENDERER.lower() if CONFIG.RENDERER else ""  # Set engine to empty string if None
    css_url = static("issues/issues.css")
    js_url = static(f"issues/issues{suffix}.js")

    if CONFIG.RENDERER in ["html2canvas", "dom-to-image"]:
        renderer_url_tag = f'<script src="{static(f"issues/{CONFIG.RENDERER}{suffix}.js")}"></script>'
    elif CONFIG.RENDERER in [None, ""]:
        renderer_url_tag = ""  # No renderer script if RENDERER is None
    else:
        raise ValueError("Invalid value for RENDERER in django-issues configuration. Check your settings")

    html = f"""
<link rel="stylesheet" href="{css_url}">
{renderer_url_tag}
<script src="{static_url}issues/axios{suffix}.js"></script>
<script id="django-issues-script" src="{js_url}" data-engine="{engine}" data-url="{url}"></script>
"""
    return mark_safe(html)  # noqa: S308



import json
from typing import TYPE_CHECKING, Any

from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from django.views import View

from .backends import get_backend
from .forms import IssueForm

if TYPE_CHECKING:
    from .types import AuthenticatedHttpRequest


class IssueAPIView(View):
    def get(self, request: "AuthenticatedHttpRequest", *args: Any, **kwargs: Any) -> HttpResponse:
        backend = get_backend(request)
        form = IssueForm(backend=backend)
        return render(request, "issues/issue_form.html", {"form": form})

    def post(self, request: "AuthenticatedHttpRequest", *args: Any, **kwargs: Any) -> JsonResponse:
        try:
            data = json.loads(request.body)
        except json.JSONDecodeError:
            return JsonResponse({"success": False, "errors": "Invalid JSON payload."}, status=400)

        backend = get_backend(request)
        form = IssueForm(data, backend=backend)
        if form.is_valid():
            backend.create_ticket(form.cleaned_data)  # type: ignore[arg-type]
            return JsonResponse({"success": True, "message": "Ticket created successfully!"})
        return JsonResponse({"success": False, "errors": form.errors, "message": "Please correct the errors below."})
